/*
hreorg.c Version 0.1.0  Hash reorganize database
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

/* read the database in physical sequence */
/* one record at a time maintaining position */
/* in hshhdr */
/* hshhdr->ofst   points to the block */

#include <bthash.h>

int hreorg(hashfmt *hshhdr, int totrcds,
   double freespc)
   {
   int rslt;
   DWORD filesize;
   int rcds;
   int adjadr;
   int esttot;
   int newtot;
   int tmptot;
   int opnerr;
   int rwflag;
   double dbltot;
   int position;
   BOOL   bResult;
   DWORD  msgerr;
   LPVOID lpMsgBuf;
   hashfmt *tmphdr;
   unsigned char *rndSeed;
   unsigned char *p,*q;
   unsigned char tmpdbname[32];
   unsigned char tmphdbn[32];
   unsigned char dbname[256];
   hshhdr->reason = 0;

   filesize = GetFileSize(hshhdr->hndl, NULL);
   if (filesize == UINT_MAX)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,           /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,              /* msglen */
         NULL);          /* insertion values */
      fprintf(stderr,"hreorg: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"hreorg: GetFileSize "
         "error on %s\n",
	 hshhdr->dbname);
      hshhdr->reason = SYS_ERR + msgerr;
      return(SYS_ERR + msgerr);
      } /* GetFileSize err */

   adjadr = filesize - HDRLEN;
   rcds   = adjadr / hshhdr->rcdnode;
   if (0 && totrcds > 0 && totrcds < rcds)
      {
      fprintf(stderr,"hreorg: total "
	 "records %d "
	 "less than actual number %d\n",
	 totrcds, rcds);
      hshhdr->reason = HASH_ERR + 2401;
      return(HASH_ERR + 2401);
      } /* totrcds too small */

   if (totrcds > 0
      && totrcds != hprime(totrcds))
      {
      fprintf(stderr,"hreorg: total "
	 "records %d is not prime\n",
	 totrcds);
      hshhdr->reason = HASH_ERR + 2402;
      return(HASH_ERR + 2402);
      } /* totrcds not prime */

   newtot = totrcds;

   if (!totrcds)
      {
      dbltot = (double) rcds
	 * (1.0 + (freespc / 100.0));
      esttot = (int) (dbltot + 0.999999);
      newtot = hprime(esttot);
      tmptot = esttot;
      while (newtot < esttot)
	 {
	 tmptot += 2;
	 newtot = hprime(tmptot);
	 } /* get smallest prime larger than rcds */
      } /* calculate new totrcds */
   tmphdr = (hashfmt *) hopent(newtot,
      hshhdr->rcdlen,
      hshhdr->keylen,
      hshhdr->keyofst,
      &opnerr,
      NULL);         /* default seed */
   if (tmphdr == NULL)
      {
      fprintf(stderr,"hreorg: hopent error "
	 "%d\n", opnerr);
      hshhdr->reason = opnerr;
      return(opnerr);
      } /* hopent err */
   position = SWPFRST;
   rslt = 0;
   while (!rslt)
      {
      rslt = hsweep(hshhdr,position,NULL);
      if (!rslt && !hshhdr->reason)
	 {
	 rslt = hisrt(tmphdr,hshhdr->rcd);
	 if (rslt)
	    {
	    fprintf(stderr,"hreorg: "
	       "hisrt error %d reason %d\n",
	       rslt, tmphdr->reason);
            hshhdr->reason = tmphdr->reason;
            return(tmphdr->reason);
	    } /* insert err */
	 } /* if good sweep */
      position = SWPNEXT;
      } /* for each record in database */
   if (rslt != 1)
      {
      fprintf(stderr,"hreorg: "
         "hsweep error %d reason %d\n",
         rslt, hshhdr->reason);
      return(hshhdr->reason);
      } /* if not end of file from sweep */
   rndSeed = (unsigned char *) rndinit();
   p = (unsigned char *) tmpdbname;
   q = (unsigned char *) p + 8;
   while (p < q)
      {
      *p++ = rndnum(26,rndSeed) + 'A';
      } /* for each char in temp dbname */
   *p = '\0';
   strcat((char *) tmpdbname,".HSH");
   strcpy((char *) dbname,hshhdr->dbname);
   strcpy((char *) tmphdbn,tmphdr->dbname);
   rwflag = hshhdr->rwflag;
   hclose(tmphdr);
   hclose(hshhdr);
   /* rename the database name */
   /* to a temporary name */
   bResult = MoveFile((char *) dbname,
      (char *) tmpdbname);
   if (!bResult)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,           /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,              /* msglen */
         NULL);          /* insertion values */
      fprintf(stderr,"hreorg: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"hreorg: error "
         "renaming %s to %s\n",
         dbname, tmpdbname);
      hshhdr->reason = SYS_ERR + msgerr;
      return(SYS_ERR + msgerr);
      } /* MoveFile failed */
   /* rename the reorganized database */
   /* to the original database name */
   bResult = MoveFile((char *) tmphdbn,
      (char *) dbname);
   if (!bResult)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,           /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,              /* msglen */
         NULL);          /* insertion values */
      fprintf(stderr,"hreorg: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"hreorg: error "
         "renaming %s to %s\n",
         tmphdbn, dbname);
      hshhdr->reason = SYS_ERR + msgerr;
      return(SYS_ERR + msgerr);
      } /* MoveFile failed */
   /* remove the temporary name */
   /* to which the original database */
   /* was renamed */
   bResult = DeleteFile((char *) tmpdbname);
   if (!bResult)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,           /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,              /* msglen */
         NULL);          /* insertion values */
      fprintf(stderr,"hreorg: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"hreorg: error "
         "removing %s\n",
	 tmpdbname);
      hshhdr->reason = SYS_ERR + msgerr;
      return(SYS_ERR + msgerr);
      } /* DeleteFile failed */

   /* The database is closed. */
   /* The calling program has to call hopen */
   /* before using the database again. */

   return(0);
   } /* hreorg */
